#!/usr/bin/ruby
# cuttle, an http proxy
# $Id: cuttle.rb,v 1.13 2004/02/22 19:34:51 ak Exp $
#
# (c) 2004 Andreas Krennmair <ak@synflood.at>
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.


# configuration variables
port = 8080

# end configuration

require 'socket'
require 'thread'
require 'log4r'

$threads = [ ]

$chunk_size = 1024 # size in which cuttle should read and write
$cleanup_time = 5 # time to wait between thread clean-up

Log4r::Logger.root.level = Log4r::INFO
$logger = Log4r::Logger.new("cuttle")
Log4r::StderrOutputter.new('console')
$logger.add('console')

class Cuttle

  class URLParser

    def initialize(url)
      @valid = true
      if (url.index(':') == nil) then # the ':' of <protocol>://...
        $logger.debug "URLParser: no ':' -> invalid"
        @valid = false
        return
      end
      @protocol = url[0..url.index(':')-1].downcase
      $logger.debug "URLParser: protocol is #{@protocol}"
      if url.length <= url.index(':')+2 or url[url.index(':')+1] != '/'[0] or url[url.index(':')+2] != '/'[0] then
        $logger.debug "#{url[url.index(':')+1]} #{url[url.index(':')+2]}"
        $logger.debug "URLParser: didn't find '//' -> invalid"
        @valid = false
        return
      end
      index = url.index(':')+3
      url2 = url[index..url.length-1]
      if url2[0] == '['[0] then # handling IPv6 addresses
        # not as nice as the V6 code. damn ':' notation for v6 addresses
        # should be thoroughly checked and redone soon
        if url2.index('/') == nil then
          @url = "/"
        else
          @url = url2[url2.index('/')..url2.length-1]
        end
        if url2.index(']:') == nil then
          @port = 80
        else
          @port = url2[url2.index(']:')+2..url2.index('/')-1]
        end
        @hostname = url2[1..url2.index(']')-1] # actually, it's the IP, but who cares?
      else # no [ -> handle as hostname or v4 address
        if url2.index('/') == nil then
          if url2.index(':') == nil then
            @hostname = url2
            @port = 80
            @url = "/"
          else
            x = url2.split(/:/)
            @hostname = x[0]
            @port = x[1]
            @url = "/"
          end
        else
          index2 = url2.index('/')
          url3 = url2[0..index2-1]
          @url = url2[index2..url2.length-1]
          if url3.index(':') == nil then
            @hostname = url3
            @port = 80
          else
            x = url3.split(/:/)
            @hostname = x[0]
            @port = x[1]
          end
        end
      end
      $logger.debug "URLParser.init: valid = #{@valid} hostname = #{@hostname} port = #{@port} url = #{@url}"
    end

    def isValid?
      @valid
    end

    def getHostname
      @hostname
    end

    def getPort
      @port
    end

    def getUrl
      @url
    end

    def to_s
      "#{@protocol}://#{@hostname}:#{@port}#{@url}"
    end

    def getProtocol
      @protocol
    end

  end

  ERR_ILLEGAL_REQUEST = 1
  ERR_NO_CONTENT_LEN = 2
  ERR_UNRESOLVABLE = 3
  ERR_CONN_FAILED = 4
  ERR_INVALID_URL = 5
  ERR_POST_PROBLEM = 6

  def initialize(port)
    $logger.debug "Initializing Cuttle (port = #{port})"
    @sessions = [ ]
    @port = port
  end


  def sendErrorPage(session,request,page,http_version)
    $logger.debug "entering sendErrorPage"
    if page == ERR_ILLEGAL_REQUEST then
      session.print "#{http_version} 400 Bad Request\r\n\r\n<html><body>Illegal request <tt>#{request}</tt><br></body></html>\r\n"
    elsif page == ERR_NO_CONTENT_LEN then
      session.print "#{http_version} 411 Length Required\r\n\r\n<html><body>You didn't send a Content-Length header with your <tt>#{request}</tt> request.<br></body></html>\r\n"
    elsif page == ERR_UNRESOLVABLE then
      session.print "#{http_version} 503 Service Unavailable\r\n\r\n<html><body>Unable to resolve <tt>#{request}</tt><br></body></html>\r\n"
    elsif page == ERR_CONN_FAILED then
      session.print "#{http_version} 503 Service Unavailable\r\n\r\n<html><body>Connection to <tt>#{request}</tt> failed.<br></body></html>\r\n"
    elsif page == ERR_INVALID_URL then
      session.print "#{http_version} 400 Bad Request\r\n\r\n<html><body>Invalid URL <tt>#{request}</tt>.<br></body></html>\r\n"
    elsif page == ERR_POST_PROBLEM then
      session.print "#{http_version} 503 Service Unavailable\r\n\r\n<html><body>Problem with <tt>POST #{request}</tt><br></body></html>\r\n"
    else
      session.print "#{http_version} 400 Bad Request\r\n\r\n<html><body>Unknown error</body></html>\r\n"
    end
    $logger.debug "leaving sendErrorPage"
  end

  def isValidRequestMethod?(m)
    x = [ "GET", "POST", "HEAD" ]
    x.each do |method|
      return true if (method == m)
    end
    return false
  end

  def getHeaderValue(headers,header_name)
    $logger.debug "Entering getHeaderValue"
    re = Regexp.new("^#{header_name}: ",Regexp::IGNORECASE)
    headers.each do |h|
      if re.match(h) then
        value = h[h.index(':')+1..h.length-1].chomp("\n").chomp("\r")
        $logger.debug "`#{h}' matches `#{header_name}', value is `#{value}'"
        return value
      end
    end
    nil
  end

  def resolveHostname(hostname)
    $logger.debug "Entering resolveHostname"
    res_result = nil
    begin
      res_result = TCPSocket.gethostbyname(hostname)
    rescue
      res_result = nil
      $logger.debug "gethostbyname: #{$!}"
    end
    $logger.debug "After gethostbyname"
    if res_result then
      res_result[3] # res_result contains canonical name, alias array, address type, IP
    else
      nil
    end
  end

  def doPost(session,ip,port,uri,http_version,headers,post_content)
    $logger.debug "Entering doPost #{ip}:#{port}#{uri}"
    t = TCPSocket.new(ip,port)
    if not t then
      sendErrorPage(session,"#{ip}:#{port}",ERR_CONN_FAILED,http_version)
    else
      $logger.debug "Established connection"
      t.print "POST #{uri} #{http_version}\r\n"
      headers.each { |h| t.print "#{h}\r\n" }
      t.print "\r\n"
      t.print "#{post_content}"
      while answer = t.read($chunk_size) do
        $logger.debug "doPost: read #{answer.length} bytes"
        begin
          session.write(answer)
        rescue
          $logger.debug "session.write failed: #{$!}"
          return
        end
        session.flush
        $logger.debug "doPost: wrote #{answer.length} bytes"
      end
    end
  end

  def doOtherMethod(session,ip,port,uri,method,http_version,headers)
    t = nil
    begin
      t = TCPSocket.new(ip,port)
    rescue
      $logger.info "Problem connecting to [#{ip}]:#{port}: #{$!}"
    end
    if not t then
      sendErrorPage(session,"#{ip}:#{port}",ERR_CONN_FAILED,http_version)
    else
      $logger.debug "doOtherMethod(#{ip}:#{port}): #{method.upcase} #{uri} #{http_version}"
      t.print "#{method.upcase} #{uri} #{http_version}\r\n"
      headers.each { |h| t.print "#{h}\r\n" ; $logger.debug "header: #{h}" }
      t.print "\r\n"
      while answer = t.read($chunk_size) do
        $logger.debug "doOtherMethod: read #{answer.length} bytes"
        begin
          session.write(answer)
        rescue
          $logger.debug "session.write failed: #{$!}"
          return
        end
        session.flush
        $logger.debug "doOtherMethod: wrote #{answer.length} bytes"
      end
    end
  end

  def getPageAndSend(session,headers,method,url,http_version)
    $logger.debug "Entering getPageAndSend method = #{method}"
    if method == "POST" then
      post_content = nil
      content_length = getHeaderValue(headers,"content-length")
      $logger.debug "content_length = #{content_length}"
      if content_length != nil then
        # content_length = content_length.to_i
        begin
          x = session.read(content_length.to_i)
        rescue 
          $logger.debug "read threw an exception" + $!
        end
        $logger.debug "after read"
        if x != nil then
          post_content = x
        else
          sendErrorPage(session,url.to_s,ERR_POST_PROBLEM,http_version)
          return
        end
        if ip = resolveHostname(url.getHostname) then
          doPost(session,ip,url.getPort,url.getUrl,http_version,headers,post_content)
        else
          sendErrorPage(session,url.getHostname,ERR_UNRESOLVABLE,http_version)
        end
      else
        $logger.debug "bar"
        sendErrorPage(session,"#{method} #{url.to_s} #{http_version}",ERR_NO_CONTENT_LEN,http_version)
      end
    else
      $logger.debug "Method is not POST, but #{method}"
      ip = resolveHostname(url.getHostname)
      $logger.debug "done resolving"
      if ip != nil then
        $logger.debug "resolveHostname successful"
        doOtherMethod(session,ip,url.getPort,url.getUrl,method,http_version,headers)
        $logger.debug "doOtherMethod done"
      else
        sendErrorPage(session,url.getHostname,ERR_UNRESOLVABLE,http_version)
      end
    end
  end


  def processSession(s)
    $logger.debug "processSession: entering"
    lines = []
    line = nil
    begin
      line = s.gets.chomp("\n").chomp("\r")
      if line != "" then
        lines << line
      else
        line = nil
      end
    end while line != nil
    fields = lines[0].split(/ /)
    method = fields[0]
    url = fields[1]
    http_version = fields[2]
    headers = lines[1..lines.size-1]
    $logger.info "#{url}"
    if (isValidRequestMethod?(method)) then
      $logger.debug "#{method} is a valid request method"
      parsed_url = URLParser.new(url)
      if parsed_url.isValid? then
        $logger.debug "parsed URL is valid"
        getPageAndSend(s,headers,method,parsed_url,http_version)
      else
        $logger.debug "parsed URL is invalid"
        sendErrorPage(s,url,ERR_INVALID_URL,http_version)
      end
      s.close
    else
      $logger.debug "#{method} is an invalid request method"
      sendErrorPage(s,lines[0],ERR_ILLEGAL_REQUEST,http_version)
      s.close
    end
    $logger.debug "processSession: leaving"
    Thread.self.kill
  end

  def runConnectionAcceptor
    $logger.debug "runConnectionAcceptor: entering"
    server = TCPServer.new('0.0.0.0',@port)
    $logger.debug "runConnectionAcceptor: started new TCPServer"
    while (session = server.accept) do
      $logger.debug "runConnectionAcceptor: accepted session"
      $threads << Thread.new { self.processSession(session) }
    end
    $logger.debug "runConnectionAcceptor: leaving"
  end

end

cuttleInstance = Cuttle.new(port)

acceptorThread = Thread.new { cuttleInstance.runConnectionAcceptor }

# monitor whether threads clean them up by themselves
while true do
  sleep $cleanup_time
  if Thread.list.size > 2 then
    $logger.debug "#{Thread.list.size} threads (should be 2 most of the time, otherwise we have a resource leak or a few request that are WIP, so if this goes on, restart cuttle and report a bug"
  end
end
